@@ -1,5 +1,5 @@
+import copy
 import logging
-from itertools import izip
 from operator import itemgetter
 
 from pyes import ES
@@ -18,6 +18,16 @@ DEFAULT_INDEXES = ['default']
 DEFAULT_DUMP_CURL = None
 
 
+class InvalidFieldActionError(Exception):
+    """Raise this when the field action doesn't exist"""
+    pass
+
+
+class InvalidFacetType(Exception):
+    """Raise when _type is unrecognized."""
+    pass
+
+
 def _split(s):
     if '__' in s:
         return s.rsplit('__', 1)
@@ -84,11 +94,6 @@ def get_es(hosts=None, default_indexes=None, timeout=None, dump_curl=None,
     return es
 
 
-class InvalidFieldActionError(Exception):
-    """Raise this when the field action doesn't exist"""
-    pass
-
-
 def _process_filters(filters):
     rv = []
     for f in filters:
@@ -102,6 +107,8 @@ def _process_filters(filters):
                 rv.append({'or':_process_filters(val.items())})
             elif field_action is None:
                 rv.append({'term': {key: val}})
+            elif field_action in ('startswith', 'prefix'):
+                rv.append({'prefix': {key: val}})
             elif field_action == 'in':
                 rv.append({'in': {key: val}})
             elif field_action in ('gt', 'gte', 'lt', 'lte'):
@@ -153,18 +160,23 @@ class F(object):
         objects combined with the connector `conn`.
         """
         f = F()
+
+        self_filters = copy.deepcopy(self.filters)
+        other_filters = copy.deepcopy(other.filters)
+
         if not self.filters:
-            f.filters = other.filters
+            f.filters = other_filters
         elif not other.filters:
-            f.filters = self.filters
+            f.filters = self_filters
         elif conn in self.filters:
-            f.filters = self.filters
-            f.filters[conn].append(other.filters)
+            f.filters = self_filters
+            f.filters[conn].append(other_filters)
         elif conn in other.filters:
-            f.filters = other.filters
-            f.filters[conn].append(self.filters)
+            f.filters = other_filters
+            f.filters[conn].append(self_filters)
         else:
-            f.filters = {conn: [self.filters, other.filters]}
+            f.filters = {conn: [self_filters, other_filters]}
+
         return f
 
     def __or__(self, other):
@@ -175,11 +187,13 @@ class F(object):
 
     def __invert__(self):
         f = F()
-        if (len(self.filters) < 2 and
-           'not' in self.filters and 'filter' in self.filters['not']):
-            f.filters = self.filters['not']['filter']
+        self_filters = copy.deepcopy(self.filters)
+        if (len(self_filters) < 2
+            and 'not' in self_filters
+            and 'filter' in self_filters['not']):
+            f.filters = self_filters['not']['filter']
         else:
-            f.filters = {'not': {'filter': self.filters}}
+            f.filters = {'not': {'filter': self_filters}}
         return f
 
 
@@ -193,10 +207,8 @@ def _boosted_value(name, action, key, value, boost):
         # Note: Most queries use 'value' for the key name except Text
         # queries which use 'query'. So we have to do some switcheroo
         # for that.
-        return {
-            name: {
-                'boost': boost,
-                'query' if action == 'text' else 'value': value}}
+        value_key = 'query' if action in ['text', 'text_phrase'] else 'value'
+        return {name: {'boost': boost, value_key: value}}
     return {name: value}
 
 
@@ -324,12 +336,21 @@ class S(object):
     def values_list(self, *fields):
         """
         Return a new S instance that returns ListSearchResults.
+
+        :arg fields: the list of fields to have in the results.
+            By default this is at least ``['id']``.
+
         """
         return self._clone(next_step=('values_list', fields))
 
     def values_dict(self, *fields):
         """
         Return a new S instance that returns DictSearchResults.
+
+        :arg fields: the list of fields to have in the results.
+            By default, this won't specify fields and thus ES
+            will return everything.
+
         """
         return self._clone(next_step=('values_dict', fields))
 
@@ -453,7 +474,8 @@ class S(object):
         filters = []
         queries = []
         sort = []
-        fields = set(['id'])
+        dict_fields = set()
+        list_fields = set()
         facets = {}
         facets_raw = {}
         demote = None
@@ -470,13 +492,16 @@ class S(object):
                     else:
                         sort.append(key)
             elif action == 'values_list':
-                fields |= set(value)
+                if not value:
+                    list_fields = set()
+                else:
+                    list_fields |= set(value)
                 as_list, as_dict = True, False
             elif action == 'values_dict':
                 if not value:
-                    fields = set()
+                    dict_fields = set()
                 else:
-                    fields |= set(value)
+                    dict_fields |= set(value)
                 as_list, as_dict = False, True
             elif action == 'explain':
                 explain = value
@@ -525,8 +550,12 @@ class S(object):
                     }
                 }
 
-        if fields:
-            qs['fields'] = list(fields)
+        if as_list and list_fields:
+            fields = qs['fields'] = list(list_fields)
+        elif as_dict and dict_fields:
+            fields = qs['fields'] = list(dict_fields)
+        else:
+            fields = set()
 
         if facets:
             qs['facets'] = facets
@@ -616,7 +645,7 @@ class S(object):
             hits = self.raw()
             if self.as_list:
                 ResultClass = ListSearchResults
-            elif self.as_dict or self.type is None:
+            elif self.as_dict:
                 ResultClass = DictSearchResults
             else:
                 ResultClass = ObjectSearchResults
@@ -624,6 +653,7 @@ class S(object):
         return self._results_cache
 
     def get_es(self, default_builder=get_es):
+        """Returns the ES object to use."""
         # The last one overrides earlier ones.
         for action, value in reversed(self.steps):
             if action == 'es_builder':
@@ -635,15 +665,28 @@ class S(object):
         return default_builder()
 
     def get_indexes(self, default_indexes=DEFAULT_INDEXES):
+        """Returns the list of indexes to act on."""
         for action, value in reversed(self.steps):
             if action == 'indexes':
-                return value
+                return list(value)
+
+        if self.type is not None:
+            indexes = self.type.get_index()
+            if isinstance(indexes, basestring):
+                indexes = [indexes]
+            return indexes
+
         return default_indexes
 
     def get_doctypes(self, default_doctypes=DEFAULT_DOCTYPES):
+        """Returns the list of doctypes to use."""
         for action, value in reversed(self.steps):
             if action == 'doctypes':
-                return value
+                return list(value)
+
+        if self.type is not None:
+            return [self.type.get_mapping_type_name()]
+
         return default_doctypes
 
     def raw(self):
@@ -653,11 +696,9 @@ class S(object):
         """
         qs = self._build_query()
         es = self.get_es()
-        try:
-            hits = es.search(qs, self.get_indexes(), self.get_doctypes())
-        except Exception:
-            log.error(qs)
-            raise
+
+        hits = es.search(qs, self.get_indexes(), self.get_doctypes())
+
         log.debug('[%s] %s' % (hits['took'], qs))
         return hits
 
@@ -674,6 +715,12 @@ class S(object):
                 facets[key] = [v for v in val['terms']]
             elif val['_type'] == 'range':
                 facets[key] = [v for v in val['ranges']]
+            elif val['_type'] == 'date_histogram':
+                facets[key] = [v for v in val['entries']]
+            else:
+                raise InvalidFacetType(
+                    'Facet _type "%s". key "%s" val "%r"' %
+                    (val['_type'], key, val))
         return facets
 
 
@@ -778,12 +825,9 @@ class MLT(object):
         if self.s:
             kwargs['body'] = self.s._build_query()
 
-        try:
-            hits = es._send_request('GET', path, **kwargs)
-            log.debug(hits)
-        except Exception:
-            log.error(kwargs)
-            raise
+        hits = es._send_request('GET', path, **kwargs)
+        log.debug(hits)
+
         log.debug('[%s] %s' % (hits['took'], kwargs))
         return hits
 
@@ -801,11 +845,11 @@ class MLT(object):
 class SearchResults(object):
     def __init__(self, type, results, fields):
         self.type = type
-        self.took = results['took']
-        self.count = results['hits']['total']
+        self.took = results.get('took', 0)
+        self.count = results.get('hits', {}).get('total', 0)
         self.results = results
         self.fields = fields
-        self.set_objects(results['hits']['hits'])
+        self.set_objects(results.get('hits', {}).get('hits', []))
 
     def set_objects(self, hits):
         raise NotImplementedError()
@@ -828,7 +872,7 @@ class TupleResult(tuple):
 class DictSearchResults(SearchResults):
     def set_objects(self, hits):
         key = 'fields' if self.fields else '_source'
-        self.objects = [_decorate_with_metadata(DictResult(r[key]), r)
+        self.objects = [decorate_with_metadata(DictResult(r[key]), r)
                         for r in hits]
 
 
@@ -846,25 +890,39 @@ class ListSearchResults(SearchResults):
                 objs = [((obj,), r) for obj, r in objs]
         else:
             objs = [(r['_source'].values(), r) for r in hits]
-        self.objects = [_decorate_with_metadata(TupleResult(obj), r)
+        self.objects = [decorate_with_metadata(TupleResult(obj), r)
                         for obj, r in objs]
 
 
+def _convert_results_to_dict(r):
+    """Takes a results from ES and returns fields."""
+    if 'fields' in r:
+        return r['fields']
+    if '_source' in r:
+        return r['_source']
+    return {'id': r['_id']}
+
+
 class ObjectSearchResults(SearchResults):
     def set_objects(self, hits):
-        self.ids = [int(r['_id']) for r in hits]
-        self.objects = self.type.objects.filter(id__in=self.ids)
+        mapping_type = (self.type if self.type is not None
+                        else DefaultMappingType)
+        self.objects = [
+            decorate_with_metadata(
+                mapping_type.from_results(_convert_results_to_dict(r)),
+                r)
+            for r in hits]
 
     def __iter__(self):
-        objs = dict((obj.id, obj) for obj in self.objects)
-        return (_decorate_with_metadata(objs[id], r)
-                for id, r in
-                izip(self.ids, self.results['hits']['hits'])
-                if id in objs)
+        return self.objects.__iter__()
 
 
-def _decorate_with_metadata(obj, hit):
+def decorate_with_metadata(obj, hit):
     """Return obj decorated with hit-scope metadata."""
+    # ES id
+    obj._id = hit.get('_id', 0)
+    # Source data
+    obj._source = hit.get('_source', {})
     # The search result score
     obj._score = hit.get('_score')
     # The document type
@@ -874,3 +932,149 @@ def _decorate_with_metadata(obj, hit):
     # Highlight bits
     obj._highlight = hit.get('highlight', {})
     return obj
+
+
+class NoModelError(Exception):
+    pass
+
+
+class MappingType(object):
+    """Base class for mapping types.
+
+    To extend this class:
+
+    1. implement ``get_index``.
+    2. implement ``get_mapping_type_name``.
+    3. if this ties back to a model, implement ``get_model``
+       and possibly also ``get_object``.
+
+    """
+    def __init__(self):
+        self._results_dict = {}
+        self._object = None
+
+    @classmethod
+    def from_results(cls, results_dict):
+        mt = cls()
+        mt._results_dict = results_dict
+        return mt
+
+    def _get_object_lazy(self):
+        if self._object:
+            return self._object
+
+        self._object = self.get_object()
+        return self._object
+
+    def get_object(self):
+        """Returns the model instance
+
+        This gets called when someone uses the ``.object`` attribute
+        which triggers lazy-loading of the object.
+
+        If this MappingType is associated with a model, then by
+        default, it calls::
+
+            self.get_model().get(id=self._id)
+
+        where ``self._id`` is the ElasticSearch document id.
+
+        Override it to do something different.
+
+        :raises cls.DoesNotExist: if the instance doesn't exist.
+            You should wrap this in a try/except block like this::
+
+                try:
+                    obj = result.object
+                except result.get_model().DoesNotExist:
+                    # exception handling here....
+
+        """
+        return self.get_model().get(id=self._id)
+
+    @classmethod
+    def get_indexes(cls):
+        """Returns the indexes to use for this mapping type.
+
+        You can specify the indexes to use for this mapping type.
+        This affects ``S`` built with this type.
+
+        By default, this is ["default"].
+
+        Override this if you want something different.
+
+        """
+        return DEFAULT_INDEXES
+
+    @classmethod
+    def get_mapping_type_name(cls):
+        """Returns the mapping type name.
+
+        You can specify the mapping type name (also sometimes called the
+        document type) with this method.
+
+        By default, this is None.
+
+        Override this if you want something different.
+
+        """
+        return DEFAULT_DOCTYPES
+
+    @classmethod
+    def get_model(cls):
+        """Return the model related to this MappingType.
+
+        This can be any class that has an instance related to this
+        Mappingtype by id.
+
+        By default, raises NoModelError.
+
+        Override this to return a class that has a ``.get(id=id)``
+        classmethod.
+
+        TODO: fix the docs.
+
+        """
+        raise NoModelError
+
+    # Simulate attribute access
+
+    def __getattr__(self, name):
+        if name in self.__dict__:
+            # We want instance/class attributes to take precedence.
+            # So if something like that exists, we raise an
+            # AttributeError and Python handles it.
+            raise AttributeError
+
+        if name == 'object':
+            # 'object' is lazy-loading. We don't do this with a
+            # property because Python sucks at properties and
+            # subclasses.
+            return self.get_object()
+
+        # If that doesn't exist, then check the results_dict.
+        if name in self._results_dict:
+            return self._results_dict[name]
+
+        raise AttributeError
+
+    # Simulate read-only container access
+
+    def __len__(self):
+        return self._results_dict.__len__()
+
+    def __getitem__(self, key):
+        return self._results_dict.__getitem__(key)
+
+    def __iter__(self):
+        return self._results_dict.__iter__()
+
+    def __reversed__(self):
+        return self._results_dict.__reversed__()
+
+    def __contains__(self, item):
+        return self._results_dict.__contains__(item)
+
+
+class DefaultMappingType(MappingType):
+    """This is the default mapping type for S."""
