@@ -1,44 +1,28 @@
+from datetime import datetime, timedelta
+
 from nose.tools import eq_
 
-from elasticutils import F, S, InvalidFieldActionError
-from elasticutils.tests import FakeModel, ElasticTestCase, facet_counts_dict
+from elasticutils import F, InvalidFieldActionError, InvalidFacetType
+from elasticutils.tests import ElasticTestCase, facet_counts_dict
 
 
-class HasDataTestCase(ElasticTestCase):
+class QueryTest(ElasticTestCase):
     @classmethod
     def setup_class(cls):
-        super(HasDataTestCase, cls).setup_class()
+        super(QueryTest, cls).setup_class()
         if cls.skip_tests:
             return
 
-        es = cls.get_es()
-        es.delete_index_if_exists(cls.index_name)
+        cls.create_index()
+        cls.index_data([
+                {'id': 1, 'foo': 'bar', 'tag': 'awesome', 'width': '2'},
+                {'id': 2, 'foo': 'bart', 'tag': 'boring', 'width': '7'},
+                {'id': 3, 'foo': 'car', 'tag': 'awesome', 'width': '5'},
+                {'id': 4, 'foo': 'duck', 'tag': 'boat', 'width': '11'},
+                {'id': 5, 'foo': 'train car', 'tag': 'awesome', 'width': '7'}
+            ])
+        cls.refresh()
 
-        data1 = FakeModel(id=1, foo='bar', tag='awesome', width='2')
-        data2 = FakeModel(id=2, foo='barf', tag='boring', width='7')
-        data3 = FakeModel(id=3, foo='car', tag='awesome', width='5')
-        data4 = FakeModel(id=4, foo='duck', tag='boat', width='11')
-        data5 = FakeModel(id=5, foo='train car', tag='awesome', width='7')
-
-        for data in (data1, data2, data3, data4, data5):
-            es.index(data.__dict__, cls.index_name, FakeModel._meta.db_table,
-                    bulk=True, id=data.id)
-        es.refresh()
-
-    @classmethod
-    def teardown_class(cls):
-        super(HasDataTestCase, cls).teardown_class()
-        if cls.skip_tests:
-            return
-
-        es = cls.get_es()
-        es.delete_index(cls.index_name)
-
-    def get_s(self):
-        return S().indexes(self.index_name).doctypes(FakeModel._meta.db_table)
-
-
-class QueryTest(HasDataTestCase):
     def test_q(self):
         eq_(len(self.get_s().query(foo='bar')), 1)
         eq_(len(self.get_s().query(foo='car')), 2)
@@ -147,7 +131,58 @@ class QueryTest(HasDataTestCase):
         with self.assertRaises(InvalidFieldActionError):
             len(self.get_s().filter(F(tag__faux='awesome')))
 
+    def test_f_mutation_with_and(self):
+        """Make sure AND doesn't mutate operands."""
+        f1 = F(fielda='tag', fieldb='boat')
+        f2 = F(fieldc='car')
+
+        f1 & f2
+        # Should only contain f1 filters.
+        eq_(sorted(f1.filters['and']),
+            sorted([{'term': {'fielda': 'tag'}},
+                    {'term': {'fieldb': 'boat'}}]))
+
+        # Should only contain f2 filters.
+        eq_(f2.filters, {'term': {'fieldc': 'car'}})
+
+    def test_f_mutation_with_or(self):
+        """Make sure OR doesn't mutate operands."""
+        f1 = F(fielda='tag', fieldb='boat')
+        f2 = F(fieldc='car')
+
+        f1 | f2
+        # Should only contain f1 filters.
+        eq_(sorted(f1.filters['and']),
+            sorted([{'term': {'fielda': 'tag'}},
+                    {'term': {'fieldb': 'boat'}}]))
+
+        # Should only contain f2 filters.
+        eq_(f2.filters, {'term': {'fieldc': 'car'}})
+
+    def test_f_mutation_with_not(self):
+        """Make sure NOT doesn't mutate operands."""
+        f1 = F(fielda='tag')
+        f2 = ~f1
+
+        # Change f2 to see if it changes f1.
+        f2.filters['not']['filter']['term']['fielda'] = 'boat'
+
+        # Should only contain f1 filters.
+        eq_(f1.filters, {'term': {'fielda': 'tag'}})
+
+        # Should only contain f2 tweaked filter.
+        eq_(f2.filters, {'not': {'filter': {'term': {'fielda': 'boat'}}}})
+
     def test_boost(self):
+        """Boosted queries shouldn't raise a SearchPhaseExecutionException."""
+        # Note: There isn't an assertion here--we just want to make
+        # sure that it runs without throwing an exception.
+        q1 = (self.get_s()
+                  .boost(foo=4.0)
+                  .query(foo='car', foo__text='car', foo__text_phrase='car'))
+        list(q1)
+
+    def test_boost_overrides(self):
         def _get_queries(search):
             # The stuff we want is buried in the search and it's in
             # the 'must' list where each item in the list is a dict
@@ -326,100 +361,50 @@ class QueryTest(HasDataTestCase):
         eq_(list(s)[0]._highlight, {})
 
 
-class ResultsTests(HasDataTestCase):
-    def test_default_results_are_dicts(self):
-        """With untyped S, return dicts."""
-        searcher = list(S().indexes(self.index_name)
-                           .doctypes(FakeModel._meta.db_table)
-                           .query(foo='bar'))
-        assert isinstance(searcher[0], dict)
+class FacetTest(ElasticTestCase):
+    def test_facet_date_histogram(self):
+        """facet_raw with date_histogram works."""
+        today = datetime.now()
+        tomorrow = today + timedelta(days=1)
 
-    def test_typed_s_returns_type(self):
-        """With typed S, return objects of type."""
-        searcher = list(S(FakeModel).indexes(self.index_name)
-                                    .doctypes(FakeModel._meta.db_table)
-                                    .query(foo='bar'))
-        assert isinstance(searcher[0], FakeModel)
+        FacetTest.create_index()
+        FacetTest.index_data([
+                {'id': 1, 'created': today},
+                {'id': 2, 'created': today},
+                {'id': 3, 'created': tomorrow},
+                {'id': 4, 'created': tomorrow},
+                {'id': 5, 'created': tomorrow},
+            ])
+        FacetTest.refresh()
 
-    def test_values_dict_results(self):
-        """With values_dict, return list of dicts."""
-        searcher = list(self.get_s().query(foo='bar').values_dict())
-        assert isinstance(searcher[0], dict)
+        qs = (self.get_s()
+              .facet_raw(created1={
+                    'date_histogram': {
+                        'interval': 'day', 'field': 'created'
+                        }
+                    }))
 
-    def test_values_list_results_no_fields(self):
-        """No values_list fields, this returns a list of tuples."""
-        searcher = list(self.get_s().query(foo='bar').values_list())
-        eq_(searcher[0], (1,))
+        # TODO: This is a mediocre test because it doesn't test the
+        # dates and it probably should.
+        facet_counts = [item['count']
+                        for item in qs.facet_counts()['created1']]
+        eq_(sorted(facet_counts), [2, 3])
 
-    def test_values_list_results(self):
-        """With values_list fields, returns list of tuples."""
-        searcher = list(self.get_s().query(foo='bar')
-                                    .values_list('foo', 'width'))
-        assert isinstance(searcher[0], tuple)
+    def test_invalid_field_type(self):
+        """Invalid _type should raise InvalidFacetType."""
+        FacetTest.create_index()
+        FacetTest.index_data([
+                {'id': 1, 'age': 30},
+                {'id': 2, 'age': 40}
+            ])
+        FacetTest.refresh()
 
-    def test_default_results_form_has_score(self):
-        """Test default results form has _score."""
-        searcher = list(self.get_s().query(foo='bar'))
-        assert hasattr(searcher[0], '_score')
-
-    def test_values_list_form_has_score(self):
-        """Test default results form has _score."""
-        searcher = list(self.get_s().query(foo='bar').values_list())
-        assert hasattr(searcher[0], '_score')
-
-    def test_values_dict_form_has_score(self):
-        """Test default results form has _score."""
-        searcher = list(self.get_s().query(foo='bar').values_dict())
-        assert hasattr(searcher[0], '_score')
-
-    def test_values_dict_no_args(self):
-        """Calling values_dict() with no args fetches all fields."""
-        eq_(S().query(fld1=2)
-               .values_dict()
-               ._build_query(),
-            {"query": {"term": {"fld1": 2}}})
-
-    def test_values_list_no_args(self):
-        """Calling values() with no args fetches only id."""
-        eq_(S(FakeModel).query(fld1=2)
-                        .values_list()
-                        ._build_query(),
-            {'query':
-                 {"term": {"fld1": 2}},
-             'fields': ['id']})
-
-    def test_values_dict_id(self):
-        """Calling values_dict('id') shouldn't return the ID field twice."""
-        eq_(S(FakeModel).query(fld1=2)
-                        .values_dict('id')
-                        ._build_query(),
-            {'query':
-                 {"term": {"fld1": 2}},
-             'fields': ['id']})
-
-    def test_values_list_id(self):
-        """Calling values('id') shouldn't return the ID field twice."""
-        eq_(S(FakeModel).query(fld1=2)
-                        .values_list('id')
-                        ._build_query(),
-            {'query':
-                 {"term": {"fld1": 2}},
-             'fields': ['id']})
-
-    def test_values_dict_implicit_id(self):
-        """Calling values_dict() always fetches ID."""
-        eq_(S(FakeModel).query(fld1=2)
-                        .values_dict('thing')
-                        ._build_query(),
-            {'query':
-                 {"term": {"fld1": 2}},
-             'fields': ['thing', 'id']})
-
-    def test_values_list_implicit_id(self):
-        """Calling values() always fetches ID."""
-        eq_(S(FakeModel).query(fld1=2)
-                        .values_list('thing')
-                        ._build_query(),
-            {'query':
-                 {"term": {"fld1": 2}},
-             'fields': ['thing', 'id']})
+        # Note: This uses histogram. If we implement handling for
+        # that, then we need to pick another facet type to fail on or
+        # do the right thing and mock the test.
+        self.assertRaises(
+            InvalidFacetType,
+            lambda: (self.get_s()
+                     .facet_raw(created1={
+                        'histogram': {'field': 'age', 'interval': 10}})
+                     .facet_counts()))
